/*
 * (C) Copyright 2000
 * DENX Software Engineering
 * Wolfgang Denk, wd@denx.de
 * All rights reserved.
 *
 * $Date: $
 * $Revision: $
 */

#include <errno.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <linux/flash.h>
#include <linux/fs.h>

int get_number (const char *, unsigned long *);
void usage (void);

const char *cmdname;

int main (int argc, char *argv[])
{
	int i, fd, part;
	struct stat stb;
	flash_info_t info;
	flash_region_info_t *region;
	flash_partition_t *partition;
	erase_info_t erase;
	unsigned long offset, size, p_offset, p_size;

	cmdname = *argv;

	if ((argc != 2) && (argc != 4)) {
		usage ();
	}

	if ((fd = open (argv[1], O_RDWR)) < 0) {
		fprintf (stderr, "%s: Can't open %s: %s\n",
				 cmdname, argv[1], strerror (errno));
		return (EXIT_FAILURE);
	}

	if (fstat(fd, &stb) < 0) {
		fprintf (stderr, "%s: Can't stat %s: %s\n",
			*argv, argv[1], strerror(errno));
		return (EXIT_FAILURE);
	}

	part = MINOR(stb.st_rdev) & FLASH_PART_MASK;

	if (ioctl (fd, FLASHGETINFO, &info) < 0) {
		fprintf (stderr, "%s: FLASHGETINFO failed: %s\n",
				 cmdname, strerror (errno));
		return (EXIT_FAILURE);
	}

	size = info.num_regions * sizeof (flash_region_info_t);

	if ((region = malloc (size)) == NULL) {
		fprintf (stderr, "%s: Can't allocate %ld bytes: %s\n",
				 cmdname, size, strerror (errno));
		return (EXIT_FAILURE);
	}

	if (ioctl (fd, FLASHGETREGIONINFO, region) < 0) {
		fprintf (stderr, "%s: FLASHGETREGIONINFO failed: %s\n",
				 cmdname, strerror (errno));
		return (EXIT_FAILURE);
	}

	if (part) {	/* read partition info */
		if ((partition = malloc(sizeof(flash_partition_t)*info.num_parts))
			== NULL) {
			fprintf (stderr, "%s: Can't malloc %d bytes: %s\n",
				*argv,
				sizeof (flash_partition_t) * info.num_parts,
				strerror(errno));
			return (EXIT_FAILURE);
		}

		if (ioctl (fd, FLASHGETPARTINFO, partition) < 0) {
			fprintf (stderr, "%s: FLASHGETPARTINFO failed: %s\n",
				*argv, strerror(errno));
			return (EXIT_FAILURE);
		}
		p_offset = partition[part-1].start;
		p_size   = partition[part-1].size;
	} else {
		p_offset = 0;
		p_size   = 0;
	}


	if (argc == 2) {
	    if (part) {
		/*
		 * erase whole partition
		 */
		printf ("Erasing Partition %d (%ld kB at offset 0x%lx)\n",
			part, p_size >> 10, p_offset);
		erase.offset = 0;
		erase.size   = p_size;
		if (ioctl (fd, FLASHERASE, &erase) < 0) {
			fprintf (stderr,
				"%s: Erase failed on %s: %s\n",
				cmdname, argv[1],
				strerror (errno)
			);
			return (EXIT_FAILURE);
		}
	    } else {
		/*
		 * erase whole device
		 */
		offset = 0;
		for (i = 0; i < info.num_regions; i++) {
			int reg_size = region[i].num_sectors * region[i].sector_size;

			erase.offset = offset;
			erase.size = reg_size;
			printf ("Erasing %2d sectors (%3d kB, total %4d kB)"
				" at offset 0x%lx\n",
				region[i].num_sectors,
				region[i].sector_size >> 10,
				reg_size >> 10,
				offset
			);
			if (ioctl (fd, FLASHERASE, &erase) < 0) {
				fprintf (stderr,
					"%s: Erase sector failed on %s: %s\n",
					cmdname, argv[1],
					strerror (errno)
				);
				return (EXIT_FAILURE);
			}
			offset += reg_size;
		}
	    }
	    close (fd);
	    return (EXIT_SUCCESS);
	}

	/*
	 * Erase part of the device
	 */
	if (get_number (argv[2], &offset) < 0) {
		fprintf (stderr, "%s: bad offset: %s\n", cmdname, argv[2]);
		usage ();
	}
	if (get_number (argv[3], &size) < 0) {
		fprintf (stderr, "%s: bad size: %s\n", cmdname, argv[3]);
		usage ();
	}

	/*
	 * The Flash driver with partition support interprets
	 * erase.offset relative to the current partition.
	 *
	 * "offset" is relative to the start of the current partition,
	 * while "p_offset" relative to the start of the whole device.
	 */

	p_size -= p_offset;	/* remaining length of partititon */
	p_offset += offset;	/* offset relative to device */

	for (i = 0; i < info.num_regions; i++) {
		unsigned long reg_size = region[i].num_sectors
					 * region[i].sector_size;
		unsigned long reg_end  = region[i].offset + reg_size;

		if ((reg_end <= p_offset) ||	/* region below erase area */
		    (region[i].offset > p_offset + size)) { /* region above erase area */
			continue;
		}
		/* region contains at least part of erase area */
		if (p_offset % region[i].sector_size) {
			fprintf (stderr, "%s: offset 0x%lx not on sector boundary "
					 "(region %u: sector size=0x%x=%dkB)\n",
					 cmdname, offset, i,
					 region[i].sector_size, region[i].sector_size >> 10);

			return (EXIT_FAILURE);
		}
		erase.offset = offset;
		erase.size = reg_end - p_offset;
		if (erase.size > size) {
			erase.size = size;
		}
		if (p_size && erase.size > p_size) {
			erase.size = p_size;	/* don't delete beyond partition */
		}

		printf ("Erasing %4ld kB = %2ld sectors (%3d kB) at offset 0x%lx",
			erase.size >> 10, erase.size / region[i].sector_size,
			region[i].sector_size >> 10, offset);
		if (part) {
			printf (" in partition %d", part);
		}
		printf ("\n");
		if (ioctl (fd, FLASHERASE, &erase) < 0) {
			fprintf (stderr, "%s: Erase sector failed on %s: %s\n",
					 cmdname, argv[1], strerror (errno));
			return (EXIT_FAILURE);
		}
		p_offset += erase.size;
		offset   += erase.size;
		size -= erase.size;
		if (!size) {
			break;
		}
	}

	if (size) {
		printf ("Warning: %ld kB not erased (size bigger than %s)\n",
			size, part ? "partition" : "device");
	}

	close (fd);

	return (EXIT_SUCCESS);
}

int get_number (const char *s, unsigned long *val)
{
	unsigned long ret;
	unsigned char c = 0;

	if (*s == '0') {			/* hex or octal number */
		++s;
		/* just 0 ? */
		if (*s == '\0') {
			*val = 0;
			return (0);
		}
		/* is it hex? */
		if ((sscanf (s, "x%lx%c", &ret, &c) <= 0) &&
			(sscanf (s, "X%lx%c", &ret, &c) <= 0)) {
			/* not hex, maybe oct ? */
			if (sscanf (s, "%lo%c", &ret, &c) <= 0) {
				return (-1);
			}
		}
	} else {
		/* must be dec then */
		if (sscanf (s, "%ld%c", &ret, &c) <= 0) {
			return (-1);
		}
	}

	switch (c) {
	case 'k':
	case 'K':
		ret <<= 10;				/* kB */
		break;
	case 'm':
	case 'M':
		ret <<= 20;				/* MB */
		break;
	}

	*val = ret;
	return (0);
}

void usage ()
{
	fprintf (stderr,
		 "Usage: %s flash_device [ offset length ]\n"
		 "offset, length:   hex (0xNN), dec (NN) or oct (0NN) numbers,\n"
		 "                  plus optional 'k' or 'm' suffix"
		 " (for kB or MB, resp.)\n",
		 cmdname
	);
	exit (1);
}
